#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Our custom errors.

We create a simple taxonomy of errors: they are all FetchErrors, but some are RetryableFetchErrors.
The distinction matters to the RetryingFileFetcher, which will re-attempt a fetch only
if it is interrupted by a RetryableFetchError.

"""
import traceback
from pathlib import Path


class FetchError(Exception):
    """
    Some sort of problem has occurred, and the program must end.
    """

    def __init__(self, message):
        self.message = message

    def terminate_fetch(self):
        print(self.message)
        traceback.print_exception(type(self), self, self.__traceback__)


class RetryableFetchError(FetchError):
    """
    A problem has occurred, but one that may go away if you try
    again in a few seconds.
    """

    retries = 1

    def terminate_fetch(self):
        print(f"A retryable problem occurred, but was not resolved after {self.retries} retries.")
        print("You may succeed by running the program again.")
        super().terminate_fetch()


class FileValidationFault(FetchError):
    """
    A file has failed to validate for some reason. This is not a retryable error.
    """

    def __init__(self, message, file: Path):
        super().__init__(message)
        self.file = file

    def terminate_fetch(self):
        print(f"File {self.file} failed validation: {self.message}")


class UnsatisfiableConstraint(FetchError):
    """
    Something the user has asked for is impossible. For instance, insisting
    on an NGAS direct copy with a destination that NGAS cannot see.
    """


class NormalModeFileExistsError(FetchError):
    """
    Shadows the built-in FileExistsError, but intended to be raised by the ForceMode.
    Not a retryable error, the file will continue to be in place each time.
    """


class NgasServiceErrorException(FetchError):
    """
    Problems that occur during fetching from NGAS are covered by this exception.
    """

    def __init__(self, message, status_code, url, reason):
        super().__init__(message)
        self.status_code = status_code
        self.url = url
        self.reason = reason

    def terminate_fetch(self):
        print(f"NGAS error: {self.message} ({self.url} returned {self.status_code} with reason {self.reason})")


class LocatorServiceException(FetchError):
    """
    Problems that occur when trying to reach the locator service
    """
