#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from typing import Dict, List, Tuple

import sqlalchemy as sa
from sqlalchemy.orm import registry

from workspaces.capability.schema import (
    CAPABILITY_NAME_FK,
    Action,
    AnnounceQa,
    ExecuteWorkflow,
    QaFail,
    QaPass,
    QueueWorkflow,
    SendMessage,
    SendNotification,
)

mapper_registry = registry()

StateMachineRepr = Dict[str, Dict[Tuple[str, str], List[Action]]]


@mapper_registry.mapped
class StateMachine:
    """
    Class representation of a generic state machine
    """

    __tablename__ = "capability_state_machines"

    capability_name = sa.Column(
        "capability_name",
        sa.String,
        sa.ForeignKey(CAPABILITY_NAME_FK),
        primary_key=True,
        comment="Name of the capability this state machine belongs to",
    )
    machine_type = sa.Column("machine_type", sa.String, comment="Type of state machine that this is")
    associated_workflows = sa.Column(
        "associated_workflows",
        sa.JSON,
        comment="JSON metadata containing relevant information for state machine generation",
    )

    @abstractmethod
    def init_state_machine(self) -> StateMachineRepr:
        """
        This method is where state machine initialization will go

        :return: Dict representation of a state machine
        """
        raise NotImplementedError


@mapper_registry.mapped
class SimpleStateMachine(StateMachine):
    """
    Class representation of your basic launch-workflow state machine
    """

    def init_state_machine(self) -> StateMachineRepr:
        """
        Initialize launch-workflow state machine

        :return: Dict repr of workflow state machine
        """
        workflow_name = self.associated_workflows["workflow_name"]

        return {
            "Start": {("Queued", "capability-submitted"): [QueueWorkflow(arguments=workflow_name)]},
            "Queued": {("Executing", "resume"): [ExecuteWorkflow(arguments=workflow_name)]},
            "Executing": {
                ("Complete", "workflow-complete"): [
                    SendMessage(arguments="execution_complete"),
                    SendNotification(arguments="workflow-complete"),
                ],
                ("Error", "workflow-failed"): [
                    SendMessage(arguments="execution_error"),
                    SendNotification(arguments="workflow-failed"),
                ],
            },
        }


@mapper_registry.mapped
class SingleQAStateMachine(StateMachine):
    """
    Class representation of a single-stage QA, launch-workflow state machine

    TODO: In the future, there will be multiple types of QaPass/QaFail; that needs to be handled here
    """

    def init_state_machine(self) -> StateMachineRepr:
        """
        Initialize launch-workflow state machine with single-stage QA

        :return: Dict repr of workflow state machine
        """
        workflow_name = self.associated_workflows["workflow_name"]

        return {
            "Start": {("Queued", "capability-submitted"): [QueueWorkflow(arguments=workflow_name)]},
            "Queued": {("Executing", "resume"): [ExecuteWorkflow(arguments=workflow_name)]},
            "Executing": {
                ("Awaiting QA", "workflow-complete"): [
                    AnnounceQa(),
                ],
                ("Error", "workflow-failed"): [
                    SendMessage(arguments="execution_error"),
                    SendNotification(arguments="workflow-failed"),
                ],
            },
            "Awaiting QA": {
                ("Ingesting", "qa-pass"): [QaPass()],
                ("Failed", "qa-fail"): [QaFail(), SendMessage(arguments="execution_failed")],
                ("Failed", "qa-complete"): [],
            },
            "Ingesting": {
                ("Complete", "ingestion-complete"): [
                    SendMessage(arguments="execution_complete"),
                    SendNotification(arguments="workflow-complete"),
                ],
                ("Error", "ingestion-failed"): [
                    SendMessage(arguments="execution_error"),
                    SendNotification(arguments="workflow-failed"),
                ],
            },
        }
